;
;  This file is copyright (c) 1987 by DDI.  All rights reserved.
;
;
;  This file defines a character device driver that does nothing.
;
;  The device driver does not need an .EXE header, so use EXE2BIN after
;  compiling.
;
RETF    macro
        db      0cbh            ;force a RET FAR without a PROC FAR
        endm

fun     segment byte
        assume  cs:fun,ds:fun,es:fun,ss:fun
;
;  Set up equates so I don't have to refer to absolute offsets
;
commandloc      equ     2       ;location of command byte in request header
statusloc       equ     3       ;location of status word in request header
returnbyteloc   equ     13      ;location of the returned byte for code 5
residentofs     equ     14      ;location of the returned resident offset
residentseg     equ     16      ;location of the returned resident segment
errorstatus     equ     8103h   ;Return code for an error of unknown command
                                ;bit 15 (8000h) = error flag
                                ;bit 8 (0100h) = done flag
                                ;bits 0-7 = error code
donestatus      equ     0100h   ;as above, this simply sets the done flag
;
;  Here, I'm going to pull a sneaky trick and fake labels in another
;  segment.
;
        org     0bch
at0bch  label   word
        org     0beh
at0beh  label   word
;
;  The origin of a device driver is at 0, not 100H.  Remember, there is no
;  PSP associated with a device.
;
        org     0
;
;  Here is the device header
;
nulldev:
                dw      -1              ;ofs of link pointer
                dw      -1              ;seg of link pointer
                dw      8000h           ;attribute, character device
                dw      strategy        ;pointer to the strategy routine
                dw      interrupt       ;pointer to the interrupt routine
                db      'ADEVICE '      ;8-byte device name, upper case so
                                        ;that it can be found
;
;  Here starts the local data
;
packetofs:      dw      0               ;offset
packetseg:      dw      0               ;segment
oldint2fofs:    dw      0               ;old int 2fh offset
oldint2fseg:    dw      0               ;old int 2fh segment
ss_save:        dw      0               ;save area for SS
sp_save:        dw      0               ;save area for SP
message_1:      db      0dh,0ah
                db      'This is some initalization message.',0dh,0ah,'$'
message_2:      db      'And this is another initalization message.'
                db      0dh,0ah,'$'
                dw      128 dup (0)     ;a local stack
localstack      label   word            ;pointer to the stack
functiontable:  dw      initalize       ;code 0, init
                dw      exit_ok         ;code 1, media check
                dw      exit_ok         ;code 2, build bpb
                dw      fake_input      ;code 3, ioctl input control string
                dw      fake_input      ;code 4, input (read)
                dw      fake_nowait     ;code 5, nondestructive no-wait input
                dw      do_status       ;code 6, input status
                dw      exit_ok         ;code 7, input flush
                dw      fake_output     ;code 8, output (write)
                dw      fake_output     ;code 9, output (write), with verify
                dw      do_status       ;code 10, output status
                dw      exit_ok         ;code 11, output flush
                dw      fake_output     ;code 12, ioctl output control string
                dw      exit_ok         ;code 13, device open  (DOS 3.x)
                dw      exit_ok         ;code 14, device close  (DOS 3.x)
                dw      exit_ok         ;code 15, removable media  (DOS 3.x)
                dw      unknown         ;code 16, undefined
                dw      unknown         ;code 17, undefined
                dw      unknown         ;code 18, undefined
                dw      exit_ok         ;code 19, generic ioctl request
                                        ;  (DOS 3.2)
                dw      unknown         ;code 20, undefined
                dw      unknown         ;code 21, undefined
                dw      unknown         ;code 22, undefined
                dw      exit_ok         ;code 23, get logical device (DOS 3.2)
                dw      exit_ok         ;code 24, set logical device (DOS 3.2)
;
;  Here is the strategy routine
;
strategy:
;
;       db      0f1h                    ;This is an undefined opcode
;
;                       Since VIM will stop when it encounters an undefined
;                       opcode, this will halt VIM at the start of the
;                       strategy routine.  This is one way of finding the
;                       device driver.
;
;                       Note that this will stop VIM at EVERY call to the
;                       device driver strategy routine!
;
        mov     word ptr cs:packetofs,bx        ;save packet offset
        mov     word ptr cs:packetseg,es        ;save packet segment
        RETF                                    ;and quit
;
;  Note that a P command used when CS:IP point to the RETF instruction will
;  stop VIM at the PUSH AX instruction below, the start of the interrupt
;  routine.
;
;
;  Here is the interrupt routine
;
interrupt:
        push    ax                      ;save AX
        push    bx                      ;and BX
        push    dx                      ;and DX
        push    si                      ;and SI
        push    ds                      ;and DS
        push    es                      ;and ES
        les     bx,dword ptr cs:packetofs        ;get pointer to the packet
        mov     al,byte ptr es:[bx+commandloc]   ;get the command
        cmp     al,24                   ;legal command?
        ja      unknown                 ;nope, skip it
        cbw                             ;make into word
        mov     si,offset functiontable ;base address of jump table
        add     si,ax                   ;base + function
        add     si,ax                   ;base + 2*function (word address)
        jmp     word ptr cs:[si]        ;jump to the function
unknown:
        mov     word ptr es:[bx+statusloc],errorstatus
                        ;status is error, done, error = unknown command
        jmp     short and_quit          ;go quit
fake_nowait:
        mov     byte ptr es:[bx+returnbyteloc],0
                        ;return a 0 for character input
fake_input:
fake_output:
do_status:
exit_ok:
        mov     word ptr es:[bx+statusloc],donestatus   ;set to done
and_quit:
        pop     es                      ;restore ES
        pop     ds                      ;and DS
        pop     si                      ;and SI
        pop     dx                      ;and DX
        pop     bx                      ;and BX
        pop     ax                      ;and AX
        RETF                            ;return to calling routine
;
;  If the device driver can be successfully installed but has problems with
;  its operation, you can use the multiplex interrupt (INT 2F) to have the
;  device driver tell you where the device driver is.
;
int2fhere:
        cmp     ah,0f0h                 ;Define ourselves as multiplex 0F0h
;
;                        Remember to test and see if the multiplex is
;                        already in use if you are going to use the
;                        multiplex interrupt in a program of yours.
;                        See the DOS Technical Reference for details.
;
        jne     continue_int_2f         ;not us, let it pass
        cmp     al,0                    ;Function = 0, get installed state
                                        ;STANDARD FUNCTION REQUEST
        jne     address_check           ;no, see if address check
        mov     al,0ffh                 ;say we are installed
                                        ;STANDARD RESPONSE
        iret                            ;and return to caller
address_check:
        cmp     al,1                    ;Function = 1, get segment address
                                        ;THIS IS OUR DEFINITION
        jne     just_return             ;if not, just do nothing
        mov     ax,cs                   ;return our segment in AX
just_return:
        iret                            ;and return
continue_int_2f:
        jmp     dword ptr cs:[oldint2fofs]      ;continue the interrupt chain
;
;  Here is the Initilization routine.  This is called only ONCE per device.
;
initalize:
;
;        db     0f1h                    ;an undefined opcode
;
;                        This undefined opcode serves the same purpose as the
;                        one above, but will only stop VIM during the
;                        initalization routine, rather than at every call to
;                        the device driver.
;
        cli                             ;do not disturb
        mov     word ptr cs:ss_save,ss  ;save current SS
        mov     word ptr cs:sp_save,sp  ;and SP
        mov     ax,cs                   ;move CS
        mov     ss,ax                   ;to SS
        mov     sp,offset localstack    ;and set SP, as well
        sti                             ;interrupts are OK again
;
        mov     ds,ax                   ;set DS to the current segment
        mov     dx,offset message_1     ;DS:DX = a string to output
        mov     ah,9                    ;function = write string
        int     21h                     ;call DOS
        mov     dx,offset message_2     ;another string
        mov     ah,9                    ;fn = write string
        int     21h                     ;call DOS
;
;  This is yet another way to find the device driver.  If the initilization
;  code works correctly, interrupt VIM when you see the 1st string being
;  displayed, and put a breakpoint in the INT 21 vector to stop the next
;  DOS call.  This will break VIM when it tries to display the 2nd string,
;  and the device driver has then been found in memory.
;
        xor     ax,ax                   ;AX = 0
        mov     ds,ax                   ;DS = 0
        mov     ax,word ptr [at0bch]            ;get the offset for INT 2f
        mov     word ptr cs:oldint2fofs,ax      ;save old offset
        mov     ax,word ptr [at0beh]            ;get the segment for INT 2f
        mov     word ptr cs:oldint2fseg,ax      ;save old segment
        mov     word ptr [at0bch],offset int2fhere ;new offset
        mov     word ptr [at0beh],cs            ;new segment
;
        mov     word ptr es:[bx+residentofs],offset initalize
                                ;last resident byte+1
        mov     word ptr es:[bx+residentseg],cs ;resident segment
;
;  Note that the initalization code is thrown away after it is used
;
;  Also note:  According to the MS-DOS (2.11) documentation that accompanies
;  the Zenith Z150, the resident address is also returned in DS:DX.
;  The IBM PC-DOS Technical Reference Manual, however, does not say that
;  the resident address should be returned in DS:DX, but does say that
;  it is the responsibility of the device driver to preserve the machine
;  state.  Since the above code works under MS-DOS 2.11, the IBM scheme is
;  followed.
;
        mov     ss,word ptr cs:ss_save  ;restore SS
        mov     sp,word ptr cs:sp_save  ;and SP
        jmp     exit_ok                 ;and quit
;
;  And that's all!
;
fun     ends
        end     nulldev
